/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
                   y
                   |
                   |
                   | v7
                   +-----------------+ v6=(1,1,0)
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
        v3 +-----------------+ v2    |
           |       |         |       |                    y
           |       |         |       |                    |
           |       |         |       | v5                 |
           |    v4 +---------|-------+------> x           |
           |      /          |      /                     |
           |    /            |    /                       +-----> x
           |  /              |  /                        /
           |/                |/                         /
           +-----------------+                         /
         / v0=(0,0,1)        v1=(1,0,1)              z
       /
     /
    z
*/
public class TextureCube
{
   public static void main(String[] args)
   {
      // Create a Model object to hold the geometry.
      Model model = new Model();
      model.doBackfaceCulling = false;

      // Create the geometry for the Model.
      Vertex v0 = new Vertex(0.0, 0.0, 1.0); // front face (CCW)
      Vertex v1 = new Vertex(1.0, 0.0, 1.0);
      Vertex v2 = new Vertex(1.0, 1.0, 1.0);
      Vertex v3 = new Vertex(0.0, 1.0, 1.0);
      Vertex v4 = new Vertex(0.0, 0.0, 0.0); // back face
      Vertex v5 = new Vertex(1.0, 0.0, 0.0);
      Vertex v6 = new Vertex(1.0, 1.0, 0.0);
      Vertex v7 = new Vertex(0.0, 1.0, 0.0);

       // left face, magenta
      v0.setColor(1.0f, 1.0f, 0.0f);
      v3.setColor(1.0f, 1.0f, 0.0f);
      v4.setColor(1.0f, 1.0f, 0.0f);
      v7.setColor(1.0f, 1.0f, 0.0f);
      // right face, yellow
      v1.setColor(0.0f, 1.0f, 1.0f);
      v2.setColor(0.0f, 1.0f, 1.0f);
      v5.setColor(0.0f, 1.0f, 1.0f);
      v6.setColor(0.0f, 1.0f, 1.0f);

      // Add the geometry to the model.
      model.addVertex(v0, v1, v2, v3, v4, v5, v6, v7);

      // Create three textures for the Model.
      Texture texture0 = new Texture("../assets/textures/Yoda.ppm");
      Texture texture1 = new Texture("../assets/textures/wood.ppm");
      // Create a "procedural" texture, pixel by pixel.
      final int SIZE = 256;
      int[] pascalImage = new int[SIZE * SIZE * 3];
      for (int y = 0; y < SIZE; y++)
      {
         for (int x = 0; x < SIZE; x++)
         {
            int c = ((x|y)%255); // Pascal's Triangle like image.
            int index = (y*SIZE + x)*3;
            pascalImage[index + 0] = c;
            pascalImage[index + 1] = c;
            pascalImage[index + 2] = c;
         }
      }
      Texture texture2 = new Texture(SIZE, SIZE, pascalImage);

      // Add the textures to the model.
      model.addTexture(texture0);  // yoda
      model.addTexture(texture1);  // wood
      model.addTexture(texture2);  // pascal

      // Add texture coordinates to the model.
      model.addTextureCoord(new TexCoord(0.0, 0.0),
                            new TexCoord(1.0, 0.0),
                            new TexCoord(1.0, 1.0),
                            new TexCoord(0.0, 1.0));

      // Three textured sides of the cube.

      // front face, yoda texture
      Triangle t0 = new Triangle(model, 0, 1, 2);
      t0.setTexture( 0 );
      t0.textured = true;
      t0.addTexCoord(0);
      t0.addTexCoord(1);
      t0.addTexCoord(2);
      Triangle t1 = new Triangle(model, 0, 2, 3);
      t1.setTexture( 0 );
      t1.textured = true;
      t1.addTexCoord(0);
      t1.addTexCoord(2);
      t1.addTexCoord(3);
      model.addPrimitive(t0, t1);

      // top face, wood texture
      Triangle t2 = new Triangle(model, 3, 2, 6);
      t2.setTexture( 1 );
      t2.textured = true;
      t2.addTexCoord(0);
      t2.addTexCoord(1);
      t2.addTexCoord(2);
      Triangle t3 = new Triangle(model, 3, 6, 7);
      t3.setTexture( 1 );
      t3.textured = true;
      t3.addTexCoord(0);
      t3.addTexCoord(2);
      t3.addTexCoord(3);
      model.addPrimitive(t2, t3);

      // back face, pascal texture
      Triangle t4 = new Triangle(model, 4, 7, 6);
      t4.setTexture( 2 );
      t4.textured = true;
      t4.addTexCoord(0);
      t4.addTexCoord(1);
      t4.addTexCoord(2);
      Triangle t5 = new Triangle(model, 4, 6, 5);
      t5.setTexture( 2 );
      t5.textured = true;
      t5.addTexCoord(0);
      t5.addTexCoord(2);
      t5.addTexCoord(3);
      model.addPrimitive(t4, t5);

      // Three non-textured sides of the cube.

      // right face, yellow
      Triangle t6 = new Triangle(model, 1, 5, 6);
      Triangle t7 = new Triangle(model, 1, 6, 2);
      model.addPrimitive(t6, t7);

       // left face, magenta
      Triangle t8 = new Triangle(model, 0, 3, 7);
      Triangle t9 = new Triangle(model, 0, 7, 4);
      model.addPrimitive(t8, t9);

      // bottom face, rainbow
      Triangle t10 = new Triangle(model, 0, 4, 5);
      Triangle t11 = new Triangle(model, 0, 5, 1);
      model.addPrimitive(t10, t11);


      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the Model to the Scene.
      scene.addPosition(new Position(model));

      // Set up the camera's view frustum.
      double fov    = 35.0;
      double aspect = 1.0;
      double near   = 1.0;
      double far    = 10.0;
      scene.camera.projPerspective(fov, aspect, near, far);

      // Set up the camera's location.
      scene.camera.viewTranslate(0, 0, 4);

      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a background color.
      fb.clearFB(Color.darkGray);

      // Render our scene into the frame buffer.
      Pipeline.render(scene, fb);
      // Save the resulting image in a file.
      fb.dumpFB2File( String.format("PPM_TextureCube_Frame%02d.ppm", 0) );

      // Spin the model 360 degrees.
      for (int i = 1; i < 72; i++)
      {
         // Rotate the model by 10 degrees.
         scene.positionList.get(0).matrix2Identity();
         scene.positionList.get(0).rotateX(i*5);
         scene.positionList.get(0).rotateZ(i*5);
         // Render again.
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_TextureCube_Frame%02d.ppm", i));
      }
   }
}