/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
    This renderer takes as its input a {@link Scene} data structure
    and a {@link FrameBuffer} data structure. This renderer mutates
    the {@link FrameBuffer}'s current viewport so that it is filled
    in with the rendered image of the scene represented by the
    {@link Scene} object.
<p>
    This implements our eighteenth rendering pipeline. This pipeline
    uses prospectively correct interpolation in the rasterization stage.
    This renderer has the same seven pipeline stages as the previous one.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // Build the initial current-transformation-matrix. This includes
      // the world-to-view coordinate transformation,
      // the normalizing view-to-camera coordinate transformation,
      // and the projection transformation to homogeneous clip coordinates.
      Matrix ctm = scene.camera.projMatrix.times(
                     scene.camera.normalizeMatrix.times(
                       scene.camera.viewMatrix));

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position(scene, position2, ctm, fb);
         }
      }
   }


   /**
      Recursively renderer a {@link Position} without any debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-clip transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-clip transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 1. Apply the current model-to-clip coordinate transformation,
         for (Vertex v : position.model.vertexList)
         {
            v.times(ctm);
         }

         // 2. Do backface culling (in homogeneous coordinates).
         if (position.model.doBackfaceCulling)
         {
            List<Primitive> primitiveList = new ArrayList<>();
            for (Primitive p : position.model.primitiveList)
            {
               if (! BackfaceCulling.isBackface(p, position.model.frontFacingIsCCW))
               {
                  // Keep the primitives with the desired orientation.
                  primitiveList.add(p);
               }
            }
            position.model.primitiveList = primitiveList;
         }

         // 3. Do the primitive assembly operation.
         PrimitiveAssembly.assemble(position.model);

         // 4. Clip each primitive (in homogeneous clip space).
         List<Primitive> primitiveList = new ArrayList<>();
         for (Primitive p : position.model.primitiveList)
         {
            primitiveList.addAll( Clip.clip(p) );
         }
         position.model.primitiveList = primitiveList;

         // 5. Apply the perspective division operation to NDC.
         for (Vertex v : position.model.vertexList)
         {
            v.perspectiveDivision();
         }

         // 6. Apply viewport transformation to each visible primitive.
         // (Alternatively, we could apply viewport transformation to every vertex?)
         // Get the viewport dimensions.
         int w = fb.getWidthVP();
         int h = fb.getHeightVP();
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (Primitive p : position.model.primitiveList)
         {
            for (int index : p.indexList)
            {
               if (! transformed[index])
               {
                  transformed[index] = true;
                  Vertex v = p.model.vertexList.get( index );
                  v.x = 0.5 + w/2.001 * (v.x + 1);
                  v.y = 0.5 + h/2.001 * (v.y + 1);
                  // NOTE: Notice the 2.001 fudge factor in the last two equations.
                  // This is explained on page 142 of
                  //    "Jim Blinn's Corner: A Trip Down The Graphics Pipeline"
                  //     by Jim Blinn, 1996, Morgan Kaufmann Publishers.
               }
            }
         }

         // 7. Rasterize each visible primitive into pixels.
         for (Primitive p : position.model.primitiveList)
         {
            Rasterize.rasterize(p, fb);
         }

         // Recursively render every nested Position of this Model.
         for (Position p : position.model.nestedPositions)
         {
            if ( p.visible )
            {
               // Do a (pre-order) depth-first-traversal from this nested Position.
               render_position(scene, p, ctm, fb);
            }
         }
      }

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position(scene, p, ctm, fb);
         }
      }
   }



   /**
      Renderer the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // Build the initial current-transformation-matrix. This includes
      // the world-to-view coordinate transformation,
      // the normalizing view-to-camera coordinate transformation,
      // and the projection transformation to homogeneous clip coordinates.
      Matrix ctm = scene.camera.projMatrix.times(
                     scene.camera.normalizeMatrix.times(
                       scene.camera.viewMatrix));

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         System.out.println("=== Render Position ================================");
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position_debug(scene, position2, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("=== End Position ================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a {@link Position} with quite a bit of debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-clip transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position_debug(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      System.out.println("=== Render Model ===================================");

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // log interesting information to standard output
         logVertexList("0. Model", position.model.vertexList);

         // 1. Apply the current model-to-clip coordinate transformation,
         for (Vertex v : position.model.vertexList)
         {
            v.times(ctm);
         }

         // log interesting information to standard output
         logVertexList("1. Projected", position.model.vertexList);

         // 2. Do back faceculling (in homogeneous coordinates).
         if (position.model.doBackfaceCulling)
         {
            List<Primitive> primitiveList = new ArrayList<>();
            for (Primitive p : position.model.primitiveList)
            {
               if (! BackfaceCulling.isBackface(p, position.model.frontFacingIsCCW))
               {
                  // Keep the primitives with the desired orientation.
                  primitiveList.add(p);
               }
            }
            position.model.primitiveList = primitiveList;
         }

         // log interesting information to standard output
         logPrimitiveList("2. Culling", position.model.primitiveList);

         // 3. Do the primitive assembly operation.
         PrimitiveAssembly.assemble(position.model);

         // log interesting information to standard output
         logPrimitiveList("3. Assembly", position.model.primitiveList);

         // 4. Clip each primitive (in homogeneous clip space).
         List<Primitive> primitiveList = new ArrayList<>();
         for (Primitive p : position.model.primitiveList)
         {
            // log interesting information to standard output
            logPrimitive("4. Clipping", p);

            List<Primitive> clipped = Clip.clip(p);
            if ( 0 != clipped.size() )
            {
               for (Primitive p2 : clipped)
               {
                  // log interesting information to standard output
                  logPrimitive("4. Clipped (accept)", p2);

                  // Keep the primitives that are visible.
                  primitiveList.add(p2);
               }
            }
            else
            {
               // log interesting information to standard output
               logPrimitive("4. Clipped (reject)", p);
            }
         }
         position.model.primitiveList = primitiveList;

         // log interesting information to standard output
         logVertexList("4. Clipped", position.model.vertexList);

         // 5. Apply the perspective division operation to NDC.
         for (Vertex v : position.model.vertexList)
         {
            v.perspectiveDivision();
         }

         // log interesting information to standard output
         logVertexList("5. NDC", position.model.vertexList);

         // 6. Apply viewport transformation to each visible primitive.
         // Get the viewport dimensions.
         int w = fb.getWidthVP();
         int h = fb.getHeightVP();
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (Primitive p : position.model.primitiveList)
         {
            for (int index : p.indexList)
            {
               if (! transformed[index])
               {
                  transformed[index] = true;
                  Vertex v = p.model.vertexList.get( index );
                  v.x = 0.5 + w/2.001 * (v.x + 1);
                  v.y = 0.5 + h/2.001 * (v.y + 1);
                  // NOTE: Notice the 2.001 fudge factor in the last two equations.
                  // This is explained on page 142 of
                  //    "Jim Blinn's Corner: A Trip Down The Graphics Pipeline"
                  //     by Jim Blinn, 1996, Morgan Kaufmann Publishers.
               }
            }

            // log interesting information to standard output
            logPrimitive("6. Viewport", p);
         }

         // 7. Rasterize each visible primitive into pixels.
         for (Primitive p : position.model.primitiveList)
         {
            // log interesting information to standard output
            logPrimitive("7. Rasterize", p);

            Rasterize.rasterize(p, fb);
         }
         System.out.println();

         // Recursively render every nested Position of this Model.
         for (Position position2 : position.model.nestedPositions)
         {
            System.out.println("====== Render Model's Nested Position ================");
            if ( position2.visible )
            {
               // Do a pre-order depth-first-traversal from this nested Position.
               render_position(scene, position2, ctm, fb);
            }
            else
            {
               System.out.println("   Hidden position.");
            }
            System.out.println("====== End Model's Nested Position ===================");
         }
      }
      else
      {
         System.out.println("   Hidden model.");
      }

      System.out.println("=== End Model ===================================");

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         System.out.println("====== Render Nested Position ========================");
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position_debug(scene, p, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("====== End Nested Position ===========================");
      }
   }


   /**
      This method prints a {@link String} representation of the given
      vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Primitive} list.

      @param stage          name for the pipeline stage
      @param primitiveList  primitive list whose string representation is to be printed
   */
   private static void logPrimitiveList(String stage, List<Primitive> primitiveList)
   {
      for (Primitive p : primitiveList)
      {
         System.out.printf( stage + ": " + p.toString() );
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link Primitive}.

      @param stage  name for the pipeline stage
      @param p      {@link Primitive} whose string representation is to be printed
   */
   private static void logPrimitive(String stage, Primitive p)
   {
      System.out.print( stage + ": " + p.toString() );
      for (Integer index : p.indexList)
      {
         Vertex v = p.model.vertexList.get(index);
         System.out.printf("    index = %3d, " + v.toString(), index);
         System.out.printf("                 " + v.toStringColor());
      }
   }
}
