/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.awt.Color;

/**
   Rasterize a {@link LineSegment} into pixels in the
   {@link FrameBuffer}'s viewport.
<p>
   This rasterization algorithm is based on
<pre>
     "Fundamentals of Computer Graphics", 3rd Edition,
      by Peter Shirley, pages 163-165.
</pre>
*/
public class RasterizeLineTextured
{
   /**
      Rasterize a {@link LineSegment} into pixels in the
      {@link FrameBuffer}'s viewport.

      @param ls  {@link LineSegment} to rasterize into the {@link FrameBuffer}
      @param fb  {@link FrameBuffer} to hold rasterized pixels
   */
   public static void rasterize(LineSegment ls, FrameBuffer fb)
   {
      boolean debug = Rasterize.debug;

      // Get this line segment's texture.
      Texture tex = ls.model.textureList.get(ls.textureIndex);

      // Make local copies of several values.
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      Vertex v0 = ls.model.vertexList.get(ls.indexList.get(0));
      Vertex v1 = ls.model.vertexList.get(ls.indexList.get(1));

      int tcIndex0 = ls.tcIndexList.get(0);
      int tcIndex1 = ls.tcIndexList.get(1);

      // Get the texture coordinates for the two vertices.
      double s0 = ls.model.texCoordList.get( tcIndex0 ).s;
      double t0 = ls.model.texCoordList.get( tcIndex0 ).t;
      double s1 = ls.model.texCoordList.get( tcIndex1 ).s;
      double t1 = ls.model.texCoordList.get( tcIndex1 ).t;

      // Round the line segment's two endpoints to the nearest
      // logical pixel. This makes the algorithm a lot simpler,
      // but it can cause a slight, but noticable, shift of the
      // line segment.
      double x0 = Math.round(v0.x),  y0 = Math.round(v0.y);
      double x1 = Math.round(v1.x),  y1 = Math.round(v1.y);

      // Get the z coordinates.
      double z0 = v0.z;
      double z1 = v1.z;

      // Rasterize a degenerate line segment (a line segment
      // that projected onto a point) as a single pixel.
      if ( (x0 == x1) && (y0 == y1) )
      {
         // Get the z-buffer value currently at this pixel of the viewport.
         double zBuff = fb.getDepthVP((int)x0 - 1, h - (int)y0);
         // If the current fragment is in front of what is in the framebuffer,
         // then overwrite the data in the framebuffer.
         if ( (z0 >= z1) && (z0 > zBuff) )
         {
            // Convert texture coordinates into pixel indices.
            int xTex = (int)(s0 * (tex.width - 1));   // column index
            int yTex = (int)(t0 * (tex.height - 1));  // row index

            // Copy color data from the texture to this pixel.
            int index = (yTex * tex.width) + xTex;
            int color = tex.pixel_buffer[index];

            fb.setPixelVP((int)x0 - 1, h - (int)y0, new Color(color));
            fb.setDepthVP((int)x0 - 1, h - (int)y0, z0);
         }
         else if (z1 > zBuff)
         {
            // Convert texture coordinates into pixel indices.
            int xTex = (int)(s1 * (tex.width - 1));   // column index
            int yTex = (int)(t1 * (tex.height - 1));  // row index

            // Copy color data from the texture to this pixel.
            int index = (yTex * tex.width) + xTex;
            int color = tex.pixel_buffer[index];

            fb.setPixelVP((int)x1 - 1, h - (int)y1, new Color(color));
            fb.setDepthVP((int)x1 - 1, h - (int)y1, z1);
         }
         return;
      }

      if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
      {
         if (x1 < x0) // swap (x0, y0, z0) with (x1, y1, z1)
         {
            double tempX = x0;
            double tempY = y0;
            double tempZ = z0;
            x0 = x1;
            y0 = y1;
            z0 = z1;
            x1 = tempX;
            y1 = tempY;
            z1 = tempZ;
            // swap the texture coordinates too
            double tempS = s0;
            double tempT = t0;
            s0 = s1;
            t0 = t1;
            s1 = tempS;
            t1 = tempT;
         }

         // Compute this line segment's slope.
         double      m = (y1 - y0) / (x1 - x0);
         double slopeZ = (z1 - z0) / (x1 - x0);
         double slopeS = (s1 - s0) / (x1 - x0);
         double slopeT = (t1 - t0) / (x1 - x0);

         // In the following loop, as x moves across the logical
         // horizontal pixels, we will compute a y value for each x.
         double y = y0;

         // Rasterize this line segment.
         for (int x = (int)x0; x <= (int)x1; x++, y += m)
         {
            // Interpolate this pixel's depth between the two endpoint's depths.
            double z = z0 + slopeZ*(x - x0);

            // Get the z-buffer value currently at this pixel of the viewport.
            double zBuff = fb.getDepthVP(x - 1, h - (int)Math.round(y));

            // If the current fragment is in front of what is in the framebuffer,
            // then overwrite the data in the framebuffer.
            if (z > zBuff)
            {
               // Interpolate this pixel's texture coordinatess
               // between the two endpoint's texture coordinates.
               double s = Math.abs(s0 + slopeS*(x - x0));
               double t = Math.abs(t0 + slopeT*(x - x0));
               // We need the Math.abs() because otherwise, we sometimes get -0.0.

               s = s - Math.floor(s);  // wrap the texture in both the
               t = t - Math.floor(t);  // s and t coordinate directions
                                       // for both pos and neg texture coordinates

               // Convert (interpolated) texture coordinates into pixel indices.
               int xTex = (int)(s * (tex.width - 1));   // column index
               int yTex = (int)(t * (tex.height - 1));  // row index

               // Copy color data from the texture to this pixel.
               int index = (yTex * tex.width) + xTex;
               int color = tex.pixel_buffer[index];

               // The value of y will almost always be between
               // two vertical pixel coordinates. By rounding off
               // the value of y, we are choosing the nearest logical
               // vertical pixel coordinate.
               fb.setPixelVP(x - 1, h - (int)Math.round(y), new Color(color));
               fb.setDepthVP(x - 1, h - (int)Math.round(y), z);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, z, s, t, color);
            }

            // Advance (x,y) to the next pixel. Since delta_x = 1, we need delta_y = m.
         }
      }
      else  // abs(slope) > 1
      {
         if (y1 < y0) // swap (x0, y0, z0) with (x1, y1, z1)
         {
            double tempX = x0;
            double tempY = y0;
            double tempZ = z0;
            x0 = x1;
            y0 = y1;
            z0 = z1;
            x1 = tempX;
            y1 = tempY;
            z1 = tempZ;
            // swap the texture coordinates too
            double tempS = s0;
            double tempT = t0;
            s0 = s1;
            t0 = t1;
            s1 = tempS;
            t1 = tempT;
         }

         // Compute this line segment's slope.
         double      m = (x1 - x0) / (y1 - y0);
         double slopeZ = (z1 - z0) / (y1 - y0);
         double slopeS = (s1 - s0) / (y1 - y0);
         double slopeT = (t1 - t0) / (y1 - y0);

         // In the following loop, as y moves across the logical
         // vertical pixels, we will compute a x value for each y.
         double x = x0;

         // Rasterize this line segment.
         for (int y = (int)y0; y <= (int)y1; x += m, y++)
         {
            // Interpolate this pixel's depth between the two endpoint's depths.
            double z = z0 + slopeZ*(y - y0);

            // Get the z-buffer value currently at this pixel of the viewport.
            double zBuff = fb.getDepthVP((int)Math.round(x) - 1, h - y);

            // If the current fragment is in front of what is in the framebuffer,
            // then overwrite the data in the framebuffer.
            if (z > zBuff)
            {
               // Interpolate this pixel's texture coordinatess
               // between the two endpoint's texture coordinates.
               double s = Math.abs(s0 + slopeS*(y - y0));
               double t = Math.abs(t0 + slopeT*(y - y0));
               // We need the Math.abs() because otherwise, we sometimes get -0.0.

               s = s - Math.floor(s);  // wrap the texture in both the
               t = t - Math.floor(t);  // s and t coordinate directions
                                       // for both pos and neg texture coordinates

               // Convert (interpolated) texture coordinates into pixel indices.
               int xTex = (int)(s * (tex.width - 1));   // column index
               int yTex = (int)(t * (tex.height - 1));  // row index

               // Copy color data from the texture to this pixel.
               int index = (yTex * tex.width) + xTex;
               int color = tex.pixel_buffer[index];

               // The value of x will almost always be between
               // two horizontal pixel coordinates. By rounding off
               // the value of x, we are choosing the nearest logical
               // horizontal pixel coordinate.
               fb.setPixelVP((int)Math.round(x) - 1, h - y, new Color(color));
               fb.setDepthVP((int)Math.round(x) - 1, h - y, z);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, z, s, t, color);
            }

            // Advance (x,y) to the next pixel. Since delta_y = 1, we need delta_x = m.
         }
      }
   }


   private static void logPixel(int w, int h, int x, double y, double z,
                                              double s, double t,
                                              int color)
   {
      Color c = new Color(color);
      int r = c.getRed();
      int g = c.getGreen();
      int b = c.getBlue();
      System.out.printf("[w=%d, h=%d]  x=%d, y=%f, z=%f, r=%f, g=%f, b=%f\n",
                          w, h, x, y, z, r, g, b);
   }


   private static void logPixel(int w, int h, double x, int y, double z,
                                              double s, double t,
                                              int color)
   {
      Color c = new Color(color);
      int r = c.getRed();
      int g = c.getGreen();
      int b = c.getBlue();
      System.out.printf("[w=%d, h=%d]  x=%f, y=%d, z=%f, r=%f, g=%f, b=%f\n",
                          w, h, x, y, z, r, g, b);
   }
}
