/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.awt.Color;

/**
   Rasterize a {@link Point} into pixels in the
   {@link FrameBuffer}'s viewport.
<p>
   Use the {@link Point}'s {@link Texture} to shade each pixel.
*/
public class RasterizePointTextured
{
   /**
      Rasterize a {@link Point} into pixels in the
      {@link FrameBuffer}'s viewport using the point's
      {@link Texture}.

      @param pt  {@code Point} to rasterize into the {@code FrameBuffer}
      @param fb  {@code FrameBuffer} to hold rasterized pixels
   */
   public static void rasterize(Primitive pt, FrameBuffer fb)
   {
      boolean debug = Rasterize.debug;

      // Get this point's texture.
      Texture tex = pt.model.textureList.get(pt.textureIndex);

      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      //for (int index : pt.indexList)
      for (int i = 0; i < pt.indexList.size(); i++)
      {
         int vertexIndex   = pt.indexList.get(i);
         int texcoordIndex = pt.tcIndexList.get(i);

         Vertex v = pt.model.vertexList.get(vertexIndex);

         // Round the point's coordinates to the nearest logical pixel.
         double x = Math.round( v.x );
         double y = Math.round( v.y );

         // Get this points's depth.
         double z = v.z;

         // Get the z-buffer value currently at this pixel of the viewport.
         double zBuff = fb.getDepthVP((int)x - 1, h - (int)y);

         // If the current fragment is in front of what is in the framebuffer,
         // then overwrite the data in the framebuffer.
         if (z > zBuff)
         {
            // Get the texture coordinates for this vertex.
            TexCoord tc = pt.model.texCoordList.get(texcoordIndex);

            double s = tc.s - Math.floor(tc.s); // wrap the texture in both the
            double t = tc.t - Math.floor(tc.t); // s and t coordinate directions for
                                                // both pos and neg texture coordinates

//          s = tc.s - (int)tc.s;  // wrap the texture in both the
//          t = tc.t - (int)tc.t;  // s and t coordinate directions

            // Convert this vertex's texture coordinates into pixel indices.
            int xTex = (int)(s * (tex.width - 1));   // column index
            int yTex = (int)(t * (tex.height - 1));  // row index

            // Get color data for this pixel from the texture.
            int index = (yTex * tex.width) + xTex;
            int color = tex.pixel_buffer[index];

            fb.setPixelVP( (int)x - 1,
                           h - (int)y,
                           new Color(color) );
            fb.setDepthVP( (int)x - 1, h - (int)y, z );

            // log interesting information to standard output
            if (debug) logPixel(w, h, x, y, z, tc.s, tc.t, color);
         }
      }
   }


   private static void logPixel(int w, int h, double x, double y, double z,
                                              double s, double t,
                                              int color)
   {
      Color c = new Color(color);
      int r = c.getRed();
      int g = c.getGreen();
      int b = c.getBlue();
      System.out.printf(
         "[w=%d, h=%d]  x=%f, y=%f z=%f, <s=%f, t=%f> r=%f, g=%f, b=%f\n",
           w, h, x, y, s, t, r, g, b);
   }
}
