/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.awt.Color;

/**
   Rasterize a {@link Triangle} into pixels in the
   {@link FrameBuffer}'s viewport.
<p>
   Use the {@link Triangle}'s {@link Texture} to shade each pixel.
*/
public class RasterizeTriangleTextured
{
   /**
      Rasterize a {@link Triangle} into pixels in the
      {@link FrameBuffer}'s viewport using the triangle's
      {@link Texture}.

      @param tri  {@code Triangle} to rasterize into the {@code FrameBuffer}
      @param fb   {@code FrameBuffer} to hold rasterized pixels
   */
   public static void rasterize(Primitive tri, FrameBuffer fb)
   {
      boolean debug = Rasterize.debug;

      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      int index0 = tri.indexList.get(0);
      int index1 = tri.indexList.get(1);
      int index2 = tri.indexList.get(2);

      Vertex v0 = tri.model.vertexList.get( index0 );
      Vertex v1 = tri.model.vertexList.get( index1 );
      Vertex v2 = tri.model.vertexList.get( index2 );

      // Get the coordinates from the three vertices.
      double x0 = v0.x,  y0 = v0.y,  z0 = v0.z,  w0 = v0.w;
      double x1 = v1.x,  y1 = v1.y,  z1 = v1.z,  w1 = v1.w;
      double x2 = v2.x,  y2 = v2.y,  z2 = v2.z,  w2 = v2.w;

      int tcIndex0 = tri.tcIndexList.get(0);
      int tcIndex1 = tri.tcIndexList.get(1);
      int tcIndex2 = tri.tcIndexList.get(2);

      // Get the texture coordinates for the three vertices.
      double s0 = tri.model.texCoordList.get( tcIndex0 ).s;
      double t0 = tri.model.texCoordList.get( tcIndex0 ).t;
      double s1 = tri.model.texCoordList.get( tcIndex1 ).s;
      double t1 = tri.model.texCoordList.get( tcIndex1 ).t;
      double s2 = tri.model.texCoordList.get( tcIndex2 ).s;
      double t2 = tri.model.texCoordList.get( tcIndex2 ).t;

      // Rasterize a degenerate triangle (a triangle that
      // projected onto a line) as three line segments.
      double area = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);
      if ( 0.0001 > Math.abs(area) )
      {
         LineSegment ls1 = new LineSegment(tri.model, index0, index1);
         LineSegment ls2 = new LineSegment(tri.model, index1, index2);
         LineSegment ls3 = new LineSegment(tri.model, index2, index0);
         ls1.textured = true;
         ls2.textured = true;
         ls3.textured = true;
         ls1.textureIndex = tri.textureIndex;
         ls2.textureIndex = tri.textureIndex;
         ls3.textureIndex = tri.textureIndex;
         ls1.tcIndexList.add(tcIndex0);
         ls1.tcIndexList.add(tcIndex1);
         ls2.tcIndexList.add(tcIndex1);
         ls2.tcIndexList.add(tcIndex2);
         ls3.tcIndexList.add(tcIndex2);
         ls3.tcIndexList.add(tcIndex0);
         RasterizeLineTextured.rasterize(ls1, fb);
         RasterizeLineTextured.rasterize(ls2, fb);
         RasterizeLineTextured.rasterize(ls3, fb);
         return;
      }

      // Recall that this triangle's viewport coordinates are in the intervals
      //     x in (-0.5, vpWidth  - 0.5)
      //     y in (-0.5, vpHeight - 0.5)
      // We want to "sample" the triangle (in viewport coordinates)
      // at the pixel locations with integer coordinates between
      //     0 and vpWidth - 1
      //     0 and vpHeight - 1
      // Our first step is to find a "pixel bounding box" for the triangle.
      // Each vertex of the triangle is rounded to integer coordinates, and
      // then we find the left-most, right-most, highest, and lowest integer
      // coordinates that bound the triangle's vertices.

      // Find the greatest integer less than or equal to x0, x1, and x2.
      int xMin = (int)x0;
      if (x1 < xMin) xMin = (int)x1;
      if (x2 < xMin) xMin = (int)x2;
      if (xMin == -1) xMin++;

      // Find the greatest integer less than or equal to y0, y1, and y2.
      int yMin = (int)y0;
      if (y1 < yMin) yMin = (int)y1;
      if (y2 < yMin) yMin = (int)y2;
      if (yMin == -1) yMin++;

      // Find the least integer greater than or equal to x0, x1, and x2.
      int xMax = (int)Math.ceil(x0);
      if (x1 > xMax) xMax = (int)Math.ceil(x1);
      if (x2 > xMax) xMax = (int)Math.ceil(x2);
      if (xMax == fb.getWidthVP()) xMax--;

      // Find the least integer greater than or equal to y0, y1, and y2.
      int yMax = (int)Math.ceil(y0);
      if (y1 > yMax) yMax = (int)Math.ceil(y1);
      if (y2 > yMax) yMax = (int)Math.ceil(y2);
      if (yMax == fb.getHeightVP()) yMax--;
      //System.err.printf("xMin=% 4d, xMax=% 4d, yMin=% 4d, yMax=% 4d\n", xMin,xMax,yMin,yMax);

    //for (int y = yMin;   y <  yMax; y++) // bottom to top
      for (int y = yMax-1; y >= yMin; y--) // top to bottom
      {
         for (int x = xMin; x < xMax; x++)  /// left to right
         {
            //  f01(x,y) = (y0-y1)*(x) + (x1-x0)*(y) + x0*y1 - x1*y0
            //  f12(x,y) = (y1-y2)*(x) + (x2-x1)*(y) + x1*y2 - x2*y1
            //  f20(x,y) = (y2-y0)*(x) + (x0-x2)*(y) + x2*y0 - x0*y2

            //  alpha = f12(x,y) / f12(x0,y0)
            //  beta  = f20(x,y) / f20(x1,y1)
            //  gamma = f01(x,y) / f01(x2,y2)

            double alpha = ( (y1-y2)*(x)  + (x2-x1)*(y)  + x1*y2 - x2*y1 )
                         / ( (y1-y2)*(x0) + (x2-x1)*(y0) + x1*y2 - x2*y1 );
            double beta  = ( (y2-y0)*(x)  + (x0-x2)*(y)  + x2*y0 - x0*y2 )
                         / ( (y2-y0)*(x1) + (x0-x2)*(y1) + x2*y0 - x0*y2 );
            double gamma = ( (y0-y1)*(x)  + (x1-x0)*(y)  + x0*y1 - x1*y0 )
                         / ( (y0-y1)*(x2) + (x1-x0)*(y2) + x0*y1 - x1*y0 );

            //System.err.printf("alpha = %f, beta = %f, gamma = %f\n", alpha,beta,gamma);

            // Create a new fragment for the FrameBuffer's viewport.
            if (alpha >= 0 && beta >= 0 && gamma >= 0)
            {
               double temp0 = alpha * w1 * w2;
               double temp1 = w0 * beta * w2;
               double temp2 = w0 * w1 * gamma;
               double temp3 = 1 / (temp0 + temp1 + temp2);
               double weight0 = temp0 * temp3;
               double weight1 = temp1 * temp3;
               double weight2 = temp2 * temp3;
               // Notice that weight0 + weight1 + weight2 == 1.

               // Interpolate z-coordinate from the triangle's vertices to this fragment.
               double zCoord = (weight0 * z0) + (weight1 * z1) + (weight2 * z2);
             //double zCoord = (alpha * z0) + (beta * z1) + (gamma * z2);

               // Get the z-buffer value currently at this pixel of the viewport.
               double zBuff = fb.getDepthVP(x - 1, h - y);

               // If the current fragment is in front of the pixel in the framebuffer,
               // then overwrite the pixel in the framebuffer with this fragment.
               if (zCoord > zBuff)
               {
                  // Interpolate texture coordinates from the triangle's vertices to this fragment.
                  double s = (weight0 * s0) + (weight1 * s1) + (weight2 * s2);
                  double t = (weight0 * t0) + (weight1 * t1) + (weight2 * t2);
                //double s = (alpha * s0) + (beta * s1) + (gamma * s2);
                //double t = (alpha * t0) + (beta * t1) + (gamma * t2);

                  s = s - Math.floor(s);  // wrap the texture in both the
                  t = t - Math.floor(t);  // s and t coordinate directions
                                          // for both pos and neg texture coordinates

//                s = s - (int)s;  // wrap the texture in both the
//                t = t - (int)t;  // s and t coordinate directions

                  //System.err.printf("s = %f, t = %f\n", s, t);

                  // Get this triangle's texture.
                  Texture tex = tri.model.textureList.get(tri.textureIndex);

                  // Convert (interpolated) texture coordinates into pixel indices.
                  int xTex = (int)(s * (tex.width - 1));   // column index
                  int yTex = (int)(t * (tex.height - 1));  // row index

                  // Copy color data from the texture to this pixel.
                  int index = (yTex * tex.width) + xTex;
                  int color = tex.pixel_buffer[index];
                  fb.setPixelVP(x - 1, h - y, new Color(color), zCoord);

                  // log interesting information to standard output
                  if (debug) logPixel(w, h, x, y, zCoord, s, t, color);
               }
            }
         }
      }
   }


   private static void logPixel(int w, int h, int x, int y, double z,
                                double s, double t,
                                int color)
   {
      Color c = new Color(color);
      int r = c.getRed();
      int g = c.getGreen();
      int b = c.getBlue();
      System.out.printf(
         "[w=%d, h=%d]  x=%d, y=%d, z=%f <s=%f, t=%f> r=%f, g=%f, b=%f\n",
           w, h, x, y, z, s, t, r, g, b);
   }
}
