/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;

/**
   A {@code Point} object has a list of integers that represent
   points in space. The integers are indices into the {@link Vertex}
   list of a {@link Model} object.
*/
public class Point extends Primitive
{
   /**
      Construct a {@code Point} object with just one integer index
      into a {@link Model}'s list of {@link Vertex} objects.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Point}
      @param i      index of the {@link Vertex} of the new {@code Point}
   */
   public Point(Model model, int i)
   {
      super(model);

      indexList.add(i);
   }


   /**
      Construct a {@code Point} from an array of integer indices.

      @param model   {@link Model} containing the {@link Vertex} list for the new {@code Point}
      @param iArray  array of indices to put into the new {@code Point}
   */
   public Point(Model model, Integer... iArray)
   {
      super(model, iArray);
   }


   /**
      Construct a {@code Point} from a {@link List} of integer indices.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Point}
      @param iList  {@link List} of indices to put into the new {@code Point}
   */
   public Point(Model model, List<Integer> iList)
   {
      super(model, iList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this {@code Point} primitive.

      @param newModel  the {@link Model} that the new {@code Point} will be part of
      @return a {@code Point} object that is a deep copy of this {@code Point}
   */
   public Point makeCopy(Model newModel)
   {
      List<Integer> newIndexList = new ArrayList<>(this.indexList.size());
      for (Integer index : this.indexList)
      {
         newIndexList.add(index);
      }
      Point result = new Point(newModel, newIndexList);

      result.textured = this.textured;
      result.textureIndex = this.textureIndex;
      for (Integer index : this.tcIndexList)
      {
         result.tcIndexList.add(index);
      }

      return result;
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Point} object
   */
   @Override
   public String toString()
   {
      String result = "Point: (";
      for (Integer index : this.indexList)
      {
         result += index + ", ";
      }
      result += ")";
      if (textured)
      {
         result += "\n  with " + this.model.textureList.get(textureIndex);
         result += "tc: (";
         for (Integer index : this.tcIndexList)
         {
            result += index + ", ";
         }
         result += ")";
      }
      return result + "\n";
   }
}
