/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   A {@code Primitive} is something that we can build
   geometric shapes out of (a "graphics primitive").
<p>
   We have three gemetric primitives, {@link LineSegment},
   {@link Face}, and {@link Triangle}.
<p>
   Notice that {@link LineSegment} is a 1-dimensional
   primitive while {@link Face} and {@link Triangle} are
   2-dimensional primitives.
<p>
   Each {@code Primitive} has a reference to a {@link Model}
   object that holds a {@link List} of {@link Vertex} objects
   that determing the geometry of the model.
<p>
   Each {@code Primitive} has a list of indices into its
   {@link Model}'s {@link List} of {@link Vertex} objects.
   These are the vertices that determine the primitive's
   geometry.
<p>
   Each {@code Primitive} has an index into its {@link Model}'s
   {@link List} of {@link Texture} objects.
<p>
   Each {@code Primitive} has a list of indices into its
   {@link Model}'s {@link List} of {@link TexCoord} objects.
<p>
   If this {@code Primitive} is textured, then its {@code indexList}
   for vertices and its {@code tcIndexList} for texture coordinates
   must have the same length. Every vertex must have a texture
   coordinate.
*/
/*
   NOTE: The Primitive class could be an inner class of
         the Model class. Then each Primitive object would
         automatically have acces to the actual Vertex list
         that the Primitive is indexing into. Here, we
         have every Primitve object hold a reference to
         the Model object that it is "part of".
*/
public abstract class Primitive
{
   // A Primitive object has a list of indices to vertices in a Model.
   public List<Integer> indexList;  // the indices for this primitive
   public Model model;              // this primitive's model object
   // A Primitive object has an index into its Model's list of Textures
   // and a list of indices to texture coordinates in the Model.
   public boolean textured;
   public int textureIndex;
   public List<Integer> tcIndexList;


   /**
      Construct an empty {@code Primitive}.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Primitive}
   */
   protected Primitive(Model model)
   {
      this.model = model;
      this.indexList = new ArrayList<>();
      this.textured = false;
      this.textureIndex = 0;
      this.tcIndexList = new ArrayList<>();
   }


   /**
      Construct a {@code Primitive} from an array of integer indices.

      @param model   {@link Model} containing the {@link Vertex} list for the new {@code Primitive}
      @param iArray  array of indices to put into the new {@code Primitive}
   */
   protected Primitive(Model model, Integer... iArray)
   {
      this(model);

      for (Integer i : iArray)
      {
         indexList.add( i );
      }
   }


   /**
      Construct a {@code Primitive} from a {@link List} of integer indices.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Primitive}
      @param iList  {@code List} of indices to put into the new {@code Primitive}
   */
   protected Primitive(Model model, List<Integer> iList)
   {
      this.model = model;
      this.indexList = iList;
      this.textured = false;
      this.textureIndex = 0;
      this.tcIndexList = new ArrayList<>();
   }


   /**
      Make a deep copy of this {@code Primitive} for another {@link Model}.

      @param newModel  the {@link Model} that the new {@code Primitive} will be part of
      @return an object that is a deep copy of this {@code Primitive} object
   */
   public abstract Primitive makeCopy(Model newModel);


   /**
      Add an index at the end of this {@code Primitive}'s index {@link List}.

      @param i  vertex index to add to this {@code Primitive}
   */
   public void addVertex(int i)
   {
      indexList.add( i );  // auto boxing
   }


   /**
      Add an index at the end of this {@code Primitive}'s index {@link List}
      of {@link TexCoord}..

      @param i  texture coordinate index to add to this {@code Primitive}
   */
   public void addTexCoord(int i)
   {
      tcIndexList.add( i );  // auto boxing
   }


   /**
      Set this {@code Primitive}'s index into its {@link Model}'s
      {@link List} of {@link Texture} objects.

      @param i  new index to a {@link Texture} object for this {@code Primitive}
   */
   public void setTexture(int i)
   {
      textureIndex = i;
   }


   /**
      Give this {@code Primitive} a uniform color.
      <p>
      NOTE: This works best when this {@code Primitive}
      object does not share its {@link Vertex} objects
      with another {@code Primitive} object.

      @param c  {@link Color} for this {@code Primitive}
   */
   public void setColor(Color c)
   {
      for (Integer index : indexList)
      {
         model.vertexList.get(index).setColor(c);
      }
   }


   /**
      Give this {@code Primitive} a uniform,
      but randomly chosen, {@link Color}.
      <p>
      NOTE: This works best when this {@code Primitive}
      object does not share its {@link Vertex} objects
      with another {@code Primitive} object.
   */
   public void setColorRandom()
   {
      if (! indexList.isEmpty())
      {
         model.vertexList.get(indexList.get(0)).setColorRandom();
         setColor( model.vertexList.get(indexList.get(0)).getColor() );
      }
   }


   /**
      Give each {@link Vertex} of this {@code Primitive}
      a randomly chosen {@link Color}.
      <p>
      NOTE: This works best when this {@code Primitive}
      object does not share its {@link Vertex} objects
      with another {@code Primitive} object.
   */
   public void setColorRainbow()
   {
      for (Integer index : indexList)
      {
         model.vertexList.get(index).setColorRandom();
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Primitive} object
   */
   @Override
   public abstract String toString();
}
