/*

*/

package renderer.scene;
import  renderer.framebuffer.FrameBuffer;

import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.Color;
//import java.awt.image.DataBufferByte;
//import java.awt.image.WritableRaster;

/**

*/
public class Texture
{
   public int width;
   public int height;
   public  int[] pixel_buffer; // contains each pixel's color data
   public String name;         // used for debugging


   /**
      Create an empty {@code Texture} with the given size.

      @param width   width of the {@code Texture}
      @param height  height of the {@code Texture}
   */
   public Texture(int width, int height)
   {
      this.width = width;
      this.height = height;
      // Create the pixel buffer.
      this.pixel_buffer = new int[this.width * this.height];
      this.name = null;
   }


   /**
      Create a {@code Texture} directly from pixel data.
      <p>
      The {@code data} array is assumed to contain
      {@code 3 * width * height} ints where every
      three ints represent the RGB color values
      for a pixel of the texture.

      @param width   width of the {@code Texture}
      @param height  height of the {@code Texture}
      @param data    array holding the RGB color values for each pixel of the {@code Texture}
   */
   public Texture(int width, int height, int[] data)
   {
      this.width = width;
      this.height = height;
      this.pixel_buffer = new int[width * height];
      this.name = null;

      for (int i = 0; i < height; i++)
      {
         for (int j = 0; j < width; j++)
         {
            int r = data[3*(i*width + j) + 0];
            int g = data[3*(i*width + j) + 1];
            int b = data[3*(i*width + j) + 2];
            pixel_buffer[(height - 1 - i)*width + j] = (new Color(r, g, b)).getRGB();
         }
      }
   }


   /**
      Create a {@code Texture} from pixel data in a PPM file.

      @param filename  must name a PPM image file with magic number P6.
   */
   public Texture(String filename) // name of the ppm file to load texture data from
   {
      this.name = filename;

      // Read the pixel data in a PPM file.
      // http://stackoverflow.com/questions/2693631/read-ppm-file-and-store-it-in-an-array-coded-with-c
      FileInputStream fis = null;
      try
      {
         fis = new FileInputStream(filename);
      }
      catch (FileNotFoundException e)
      {
         e.printStackTrace(System.err);
         System.err.printf("ERROR! Could not open texture %s\n", filename);
         System.exit(-1);
      }

      try
      {
         // Read image format string "P6".
         String magicNumber = "";
         char c = (char)fis.read();
         while (c != '\n')
         {
            magicNumber += c;
            c = (char)fis.read();
         }
         if (! magicNumber.trim().startsWith("P6"))
         {
            System.err.printf("ERROR! Improper PPM number in file %s\n", filename);
            System.exit(-1);
         }

         c = (char)fis.read();
         if ( '#' == c ) // read (and discard) IrfanView comment
         {
            while (c != '\n')
            {
               c = (char)fis.read();
            }
            c = (char)fis.read();
         }

         // Read image dimensions.
         String widthDim = "";
         while (c != ' ' && c != '\n')
         {
            widthDim += c;
            c = (char)fis.read();
         }
         width = Integer.parseInt(widthDim.trim());

         String heightDim = "";
         c = (char)fis.read();
         while (c != '\n')
         {
            heightDim += c;
            c = (char)fis.read();
         }
         height = Integer.parseInt(heightDim.trim());

         System.err.printf("Texture %s: width=%d, height=%d\n\n",filename,width,height);

         // Read image rgb dimensions (which we don't use).
         c = (char)fis.read();
         while (c != '\n')
         {
            c = (char)fis.read();
         }

         // Create the pixel buffer.
         pixel_buffer = new int[this.width * this.height];

         // Create a pixel data array.
         byte[] pixelData = new byte[3];

         // Read pixel data, one pixel at a time, from the PPM file.
         for (int y = height - 1; y >= 0; y--)
         {
            for (int x = 0; x < width; x++)
            {
               if ( fis.read(pixelData, 0, 3) != 3 )
               {
                  System.err.printf("ERROR! Could not load %s\n", filename);
                  System.exit(-1);
               }
               int r = pixelData[0];
               int g = pixelData[1];
               int b = pixelData[2];
               if (r < 0) r = 256+r; // NOTE: Java doesn't have an unsigned
               if (g < 0) g = 256+g; // byte data type so we need to convert
               if (b < 0) b = 256+b; // negative bytes into postive ints.
               pixel_buffer[(y * width) + x] = (new Color(r, g, b)).getRGB();
            }
         }
         fis.close();
      }
      catch (IOException e)
      {
         e.printStackTrace(System.err);
         System.err.printf("ERROR! Could not read texture %s\n", filename);
         System.exit(-1);
      }
   }


   /**
      Create a {@code Texture} from a {@link FrameBuffer}'s view port.

      @param fb  {@link FrameBuffer} whose viewport pixel data will become a {@code Texture}
   */
   public Texture(FrameBuffer fb)
   {
      this.width  = fb.getWidthVP();
      this.height = fb.getHeightVP();
      // Create the pixel buffer.
      this.pixel_buffer = new int[this.width * this.height];
      for (int i = 0; i < height; i++)
      {
         for (int j = 0; j < width; j++)
         {
            pixel_buffer[(height - 1 - i)*width + j] = fb.getPixelVP(j,i).getRGB();
         }
      }
      this.name = null;
   }


   /**
      Create a {@code Texture} object from pixel data in an image file.

      @param filename  name of a PNG, JPG, or GIF image file
      @return {@code Texture} object created from the input file
   */
   public static Texture makeTexture(String filename)
   {
      Texture result = null;
      try
      {
         DataInputStream dis = new DataInputStream(
                                   new BufferedInputStream(
                                       new FileInputStream(filename)));

         if (Integer.toHexString(dis.readInt()).substring(0,4).equals("5036"))
         {
            result = new Texture(filename);
         }
         else
         {
            result = makeTextureNotPPM(filename);
         }
         result.name = filename;
      }
      catch(IOException e)
      {
         e.printStackTrace(System.err);
         System.err.println("File not found.");
         System.exit(-1);
      }
      return result;
   }


   /**
      Read the {@code Texture} data from a file
      using Java's {@link ImageIO} library.

      @param filename  name of a PNG, JPG, or GIF image file
      @return {@code Texture} object created from the input file
   */
   private static Texture makeTextureNotPPM(String filename)
   {
      BufferedImage img = null;

      try
      {
         img = ImageIO.read(new File(filename));
      }
      catch(IOException e)
      {
         e.printStackTrace(System.err);
         System.err.println("Cannot read image file.");
         System.exit(-1);
      }

      int width  = img.getWidth();
      int height = img.getHeight();

      Texture texture = new Texture(width, height);
      texture.name = filename;

      for (int i = 0; i < height; i++)
      {
         for (int j = 0; j < width; j++)
         {
            texture.pixel_buffer[(height - 1 - i)*width + j] = img.getRGB(j,i);
         }
      }

      System.err.printf("Texture %s: width=%d, height=%d\n\n",filename,width,height);

      return texture;
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Texture} object
   */
   @Override
   public String toString()
   {
      String result = "Texture: (width="+width+", height="+height+")";
      result += " name = " + "\"" + name + "\"";
      return result;
   }
}
